# coding=utf-8

import os
import unittest

import kgb
from djblets.testing.decorators import add_fixtures

from reviewboard import get_manual_url
from reviewboard.diffviewer.parser import DiffParserError
from reviewboard.diffviewer.testing.mixins import DiffParserTestingMixin
from reviewboard.scmtools.core import PRE_CREATION
from reviewboard.scmtools.errors import SCMError, FileNotFoundError
from reviewboard.scmtools.git import ShortSHA1Error, GitClient, GitTool
from reviewboard.scmtools.tests.testcases import SCMTestCase
from reviewboard.testing.testcase import TestCase


class GitTests(DiffParserTestingMixin, kgb.SpyAgency, SCMTestCase):
    """Unit tests for Git."""

    fixtures = ['test_scmtools']

    ssh_required_system_exes = ['git']

    def setUp(self):
        super(GitTests, self).setUp()

        self.local_repo_path = os.path.join(os.path.dirname(__file__),
                                            '..', 'testdata', 'git_repo')
        self.git_ssh_path = ('localhost:%s'
                             % self.local_repo_path.replace('\\', '/'))
        remote_repo_path = 'git@github.com:reviewboard/reviewboard.git'
        remote_repo_raw_url = ('http://github.com/api/v2/yaml/blob/show/'
                               'reviewboard/reviewboard/<revision>')

        self.repository = self.create_repository(
            name='Git test repo',
            path=self.local_repo_path,
            tool_name='Git')
        self.remote_repository = self.create_repository(
            name='Remote Git test repo',
            path=remote_repo_path,
            raw_file_url=remote_repo_raw_url,
            tool_name='Git')

        try:
            self.tool = self.repository.get_scmtool()
            self.remote_tool = self.remote_repository.get_scmtool()
        except ImportError:
            raise unittest.SkipTest('git binary not found')

    def _read_diff_fixture(self, filename, expected_num_diffs):
        """Read a diff fixture from the test data.

        Args:
            filename (unicode):
                The name of the filename in the :file:`testdata` directory.

            expected_num_diffs (int):
                The expected number of files found in the fixture.

        Returns:
            tuple:
            A 2-tuple containing:

            1. The full diff from the file.
            2. Each diff in the file.

        Raises:
            AssertionError:
                The number of diffs found did not meet the expected number.
        """
        filename = os.path.join(os.path.dirname(__file__),
                                '..', 'testdata', filename)

        with open(filename, 'rb') as f:
            full_diff = f.read()

        diffs = []
        i1 = 0

        while True:
            i2 = full_diff.find(b'diff --git', i1 + 1)

            if i2 == -1:
                break

            diffs.append(full_diff[i1:i2])
            i1 = i2

        diffs.append(full_diff[i1:])

        self.assertEqual(len(diffs), expected_num_diffs)

        return full_diff, diffs

    def _get_file_in_diff(self, diff, filenum=0):
        files = self.tool.get_parser(diff).parse()
        self.assertTrue(filenum < len(files))
        return files[filenum]

    def test_ssh(self):
        """Testing a SSH-backed git repository"""
        self._test_ssh(self.git_ssh_path)

    def test_ssh_with_site(self):
        """Testing a SSH-backed git repository with a LocalSite"""
        self._test_ssh_with_site(self.git_ssh_path)

    def test_filemode_diff(self):
        """Testing parsing filemode changes Git diff"""
        diff1 = (
            b'diff --git a/testing b/testing\n'
            b'old mode 100755\n'
            b'new mode 100644\n'
            b'index e69de29..bcae657\n'
            b'--- a/testing\n'
            b'+++ b/testing\n'
            b'@@ -0,0 +1 @@\n'
            b'+ADD\n'
        )
        diff2 = (
            b'diff --git a/testing2 b/testing2\n'
            b'old mode 100644\n'
            b'new mode 100755\n'
        )
        diff = diff1 + diff2

        # NOTE: testing2 gets skipped, due to lack of changes we can
        #       represent.
        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'testing',
            orig_file_details=b'e69de29',
            modified_filename=b'testing',
            modified_file_details=b'bcae657',
            old_unix_mode='100755',
            new_unix_mode='100644',
            insert_count=1,
            data=diff1)

    def test_filemode_with_following_diff(self):
        """Testing parsing filemode changes with following Git diff"""
        diff1 = (
            b'diff --git a/testing b/testing\n'
            b'old mode 100755\n'
            b'new mode 100644\n'
            b'index e69de29..bcae657\n'
            b'--- a/testing\n'
            b'+++ b/testing\n'
            b'@@ -0,0 +1 @@\n'
            b'+ADD\n'
        )
        diff2 = (
            b'diff --git a/testing2 b/testing2\n'
            b'old mode 100644\n'
            b'new mode 100755\n'
        )
        diff3 = (
            b'diff --git a/cfg/testcase.ini b/cfg/testcase.ini\n'
            b'index cc18ec8..5e70b73 100644\n'
            b'--- a/cfg/testcase.ini\n'
            b'+++ b/cfg/testcase.ini\n'
            b'@@ -1,6 +1,7 @@\n'
            b'+blah blah blah\n'
            b' [mysql]\n'
            b' host = localhost\n'
            b' port = 3306\n'
            b' user = user\n'
            b' pass = pass\n'
            b'-db = pyunit\n'
            b'+db = pyunit\n'
        )
        diff = diff1 + diff2 + diff3

        # NOTE: testing2 gets skipped, due to lack of changes we can
        #       represent.
        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'testing',
            orig_file_details=b'e69de29',
            modified_filename=b'testing',
            modified_file_details=b'bcae657',
            old_unix_mode='100755',
            new_unix_mode='100644',
            insert_count=1,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'cfg/testcase.ini',
            orig_file_details=b'cc18ec8',
            modified_filename=b'cfg/testcase.ini',
            modified_file_details=b'5e70b73',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=2,
            delete_count=1,
            data=diff3)

    def test_simple_diff(self):
        """Testing parsing simple Git diff"""
        diff = (
            b'diff --git a/cfg/testcase.ini b/cfg/testcase.ini\n'
            b'index cc18ec8..5e70b73 100644\n'
            b'--- a/cfg/testcase.ini\n'
            b'+++ b/cfg/testcase.ini\n'
            b'@@ -1,6 +1,7 @@\n'
            b'+blah blah blah\n'
            b' [mysql]\n'
            b' host = localhost\n'
            b' port = 3306\n'
            b' user = user\n'
            b' pass = pass\n'
            b'-db = pyunit\n'
            b'+db = pyunit\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'cfg/testcase.ini',
            orig_file_details=b'cc18ec8',
            modified_filename=b'cfg/testcase.ini',
            modified_file_details=b'5e70b73',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=2,
            delete_count=1,
            data=diff)

    def test_diff_with_unicode(self):
        """Testing parsing Git diff with unicode characters"""
        diff = (
            'diff --git a/cfg/téstcase.ini b/cfg/téstcase.ini\n'
            'index cc18ec8..5e70b73 100644\n'
            '--- a/cfg/téstcase.ini\n'
            '+++ b/cfg/téstcase.ini\n'
            '@@ -1,6 +1,7 @@\n'
            '+blah blah blah\n'
            ' [mysql]\n'
            ' hóst = localhost\n'
            ' pórt = 3306\n'
            ' user = user\n'
            ' pass = pass\n'
            '-db = pyunít\n'
            '+db = pyunít\n'
        ).encode('utf-8')

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename='cfg/téstcase.ini'.encode('utf-8'),
            orig_file_details=b'cc18ec8',
            modified_filename='cfg/téstcase.ini'.encode('utf-8'),
            modified_file_details=b'5e70b73',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=2,
            delete_count=1,
            data=diff)

    def test_diff_with_tabs_after_filename(self):
        """Testing parsing Git diffs with tabs after the filename"""
        diff = (
            b'diff --git a/README b/README\n'
            b'index 712544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/README\t\n'
            b'+++ b/README\t\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
            b'-\n'
            b'1.7.1\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'README',
            orig_file_details=b'712544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'README',
            modified_file_details=b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            delete_count=2,
            data=diff)

    def test_new_file_diff(self):
        """Testing parsing Git diff with new file"""
        diff = (
            b'diff --git a/IAMNEW b/IAMNEW\n'
            b'new file mode 100644\n'
            b'index 0000000..e69de29\n'
            b'--- /dev/null\n'
            b'+++ b/IAMNEW\n'
            b'@@ -0,0 +1,1 @@\n'
            b'+Hello\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'IAMNEW',
            orig_file_details=PRE_CREATION,
            modified_filename=b'IAMNEW',
            modified_file_details=b'e69de29',
            new_unix_mode='100644',
            insert_count=1,
            data=diff)

    def test_new_file_no_content_diff(self):
        """Testing parsing Git diff new file, no content"""
        diff = (
            b'diff --git a/newfile b/newfile\n'
            b'new file mode 100644\n'
            b'index 0000000..e69de29\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'newfile',
            orig_file_details=PRE_CREATION,
            modified_filename=b'newfile',
            modified_file_details=b'e69de29',
            new_unix_mode='100644',
            data=diff)

    def test_new_file_no_content_with_following_diff(self):
        """Testing parsing Git diff new file, no content, with following"""
        diff1 = (
            b'diff --git a/newfile b/newfile\n'
            b'new file mode 100644\n'
            b'index 0000000..e69de29\n'
        )
        diff2 = (
            b'diff --git a/cfg/testcase.ini b/cfg/testcase.ini\n'
            b'index cc18ec8..5e70b73 100644\n'
            b'--- a/cfg/testcase.ini\n'
            b'+++ b/cfg/testcase.ini\n'
            b'@@ -1,6 +1,7 @@\n'
            b'+blah blah blah\n'
            b' [mysql]\n'
            b' host = localhost\n'
            b' port = 3306\n'
            b' user = user\n'
            b' pass = pass\n'
            b'-db = pyunit\n'
            b'+db = pyunit\n'
        )
        diff = diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'newfile',
            orig_file_details=PRE_CREATION,
            modified_filename=b'newfile',
            modified_file_details=b'e69de29',
            new_unix_mode='100644',
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'cfg/testcase.ini',
            orig_file_details=b'cc18ec8',
            modified_filename=b'cfg/testcase.ini',
            modified_file_details=b'5e70b73',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=2,
            delete_count=1,
            data=diff2)

    def test_del_file_diff(self):
        """Testing parsing Git diff with deleted file"""
        diff = (
            b'diff --git a/OLDFILE b/OLDFILE\n'
            b'deleted file mode 100644\n'
            b'index 8ebcb01..0000000\n'
            b'--- a/OLDFILE\n'
            b'+++ /dev/null\n'
            b'@@ -1,1 +0,0 @@\n'
            b'-Goodbye\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'OLDFILE',
            orig_file_details=b'8ebcb01',
            modified_filename=b'OLDFILE',
            modified_file_details=b'0000000',
            old_unix_mode='100644',
            deleted=True,
            delete_count=1,
            data=diff)

    def test_del_file_no_content_diff(self):
        """Testing parsing Git diff with deleted file, no content"""
        diff = (
            b'diff --git a/empty b/empty\n'
            b'deleted file mode 100644\n'
            b'index e69de29bb2d1d6434b8b29ae775ad8c2e48c5391..'
            b'0000000000000000000000000000000000000000\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty',
            orig_file_details=b'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391',
            modified_filename=b'empty',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff)

    def test_del_file_no_content_with_following_diff(self):
        """Testing parsing Git diff with deleted file, no content, with
        following
        """
        diff1 = (
            b'diff --git a/empty b/empty\n'
            b'deleted file mode 100644\n'
            b'index e69de29bb2d1d6434b8b29ae775ad8c2e48c5391..'
            b'0000000000000000000000000000000000000000\n'
        )
        diff2 = (
            b'diff --git a/foo/bar b/foo/bar\n'
            b'index 484ba93ef5b0aed5b72af8f4e9dc4cfd10ef1a81..'
            b'0ae4095ddfe7387d405bd53bd59bbb5d861114c5 100644\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar\n'
            b'@@ -1 +1,2 @@\n'
            b'+Hello!\n'
            b'blah\n'
        )
        diff = diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'empty',
            orig_file_details=b'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391',
            modified_filename=b'empty',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'foo/bar',
            orig_file_details=b'484ba93ef5b0aed5b72af8f4e9dc4cfd10ef1a81',
            modified_filename=b'foo/bar',
            modified_file_details=b'0ae4095ddfe7387d405bd53bd59bbb5d861114c5',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            data=diff2)

    def test_binary_diff(self):
        """Testing parsing Git diff with binary"""
        diff = (
            b'diff --git a/pysvn-1.5.1.tar.gz b/pysvn-1.5.1.tar.gz\n'
            b'new file mode 100644\n'
            b'index 0000000..86b520c\n'
            b'Binary files /dev/null and b/pysvn-1.5.1.tar.gz differ\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'pysvn-1.5.1.tar.gz',
            orig_file_details=PRE_CREATION,
            modified_filename=b'pysvn-1.5.1.tar.gz',
            modified_file_details=b'86b520c',
            new_unix_mode='100644',
            binary=True,
            data=diff)

    def test_git_new_single_binary_diff(self):
        """Testing parsing Git diff with base64 binary and a new file"""
        full_diff, diffs = self._read_diff_fixture(
            'git_new_single_binary.diff',
            expected_num_diffs=2)

        parsed_files = self.tool.get_parser(full_diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'Checked.svg',
            orig_file_details=PRE_CREATION,
            modified_filename=b'Checked.svg',
            modified_file_details=b'',
            new_unix_mode='100644',
            insert_count=9,
            data=diffs[0])

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'dialog.jpg',
            orig_file_details=b'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391',
            modified_filename=b'dialog.jpg',
            modified_file_details=b'5503573346e25878d57775ed7caf88f2eb7a7d98',
            new_unix_mode='100644',
            binary=True,
            data=diffs[1])

    def test_git_new_binaries_diff(self):
        """Testing parsing Git diff with base64 binaries and new files"""
        full_diff, diffs = self._read_diff_fixture(
            'git_new_binaries.diff',
            expected_num_diffs=3)

        parsed_files = self.tool.get_parser(full_diff).parse()
        self.assertEqual(len(parsed_files), 3)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'other.png',
            orig_file_details=b'e69de29bb2d1d6434b8b29ae775ad8c2e48c5391',
            modified_filename=b'other.png',
            modified_file_details=b'fddeadc701ac6dd751b8fc70fe128bd29e54b9b0',
            new_unix_mode='100644',
            binary=True,
            data=diffs[0])

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'initial.png',
            orig_file_details=b'fddeadc701ac6dd751b8fc70fe128bd29e54b9b0',
            modified_filename=b'initial.png',
            modified_file_details=b'532716ada15dc62ddf8c59618b926f34d4727d77',
            binary=True,
            data=diffs[1])

        self.assert_parsed_diff_file(
            parsed_files[2],
            orig_filename=b'xtxt.txt',
            orig_file_details=PRE_CREATION,
            modified_filename=b'xtxt.txt',
            modified_file_details=b'',
            new_unix_mode='100644',
            insert_count=1,
            data=diffs[2])

    def test_complex_diff(self):
        """Testing parsing Git diff with existing and new files"""
        full_diff, diffs = self._read_diff_fixture(
            'git_complex.diff',
            expected_num_diffs=7)

        parsed_files = self.tool.get_parser(full_diff).parse()
        self.assertEqual(len(parsed_files), 7)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'cfg/testcase.ini',
            orig_file_details=b'5e35098',
            modified_filename=b'cfg/testcase.ini',
            modified_file_details=b'e254ef4',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=2,
            delete_count=1,
            data=diffs[0])

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'tests/models.py',
            orig_file_details=PRE_CREATION,
            modified_filename=b'tests/models.py',
            modified_file_details=b'e69de29',
            new_unix_mode='100644',
            data=diffs[1])

        self.assert_parsed_diff_file(
            parsed_files[2],
            orig_filename=b'tests/tests.py',
            orig_file_details=PRE_CREATION,
            modified_filename=b'tests/tests.py',
            modified_file_details=b'e279a06',
            new_unix_mode='100644',
            insert_count=2,
            data=diffs[2])

        self.assert_parsed_diff_file(
            parsed_files[3],
            orig_filename=b'pysvn-1.5.1.tar.gz',
            orig_file_details=PRE_CREATION,
            modified_filename=b'pysvn-1.5.1.tar.gz',
            modified_file_details=b'86b520c',
            new_unix_mode='100644',
            binary=True,
            data=diffs[3])

        self.assert_parsed_diff_file(
            parsed_files[4],
            orig_filename=b'readme',
            orig_file_details=b'5e35098',
            modified_filename=b'readme',
            modified_file_details=b'e254ef4',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            delete_count=1,
            data=diffs[4])

        self.assert_parsed_diff_file(
            parsed_files[5],
            orig_filename=b'OLDFILE',
            orig_file_details=b'8ebcb01',
            modified_filename=b'OLDFILE',
            modified_file_details=b'0000000',
            old_unix_mode='100644',
            deleted=True,
            delete_count=1,
            data=diffs[5])

        self.assert_parsed_diff_file(
            parsed_files[6],
            orig_filename=b'readme2',
            orig_file_details=b'5e43098',
            modified_filename=b'readme2',
            modified_file_details=b'e248ef4',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            delete_count=1,
            data=diffs[6])

    def test_parse_diff_with_index_range(self):
        """Testing Git diff parsing with an index range"""
        diff = (
            b'diff --git a/foo/bar b/foo/bar2\n'
            b'similarity index 88%\n'
            b'rename from foo/bar\n'
            b'rename to foo/bar2\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar2\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo/bar',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo/bar2',
            modified_file_details=b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100644',
            new_unix_mode='100644',
            moved=True,
            insert_count=1,
            delete_count=1,
            data=diff)

    def test_parse_diff_with_deleted_binary_files(self):
        """Testing Git diff parsing with deleted binary files"""
        diff1 = (
            b'diff --git a/foo.bin b/foo.bin\n'
            b'deleted file mode 100644\n'
            b'Binary file foo.bin has changed\n'
        )
        diff2 = (
            b'diff --git a/bar.bin b/bar.bin\n'
            b'deleted file mode 100644\n'
            b'Binary file bar.bin has changed\n'
        )
        diff = diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo.bin',
            orig_file_details=b'',
            modified_filename=b'foo.bin',
            modified_file_details=b'',
            old_unix_mode='100644',
            deleted=True,
            binary=True,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'bar.bin',
            orig_file_details=b'',
            modified_filename=b'bar.bin',
            modified_file_details=b'',
            old_unix_mode='100644',
            deleted=True,
            binary=True,
            data=diff2)

    def test_parse_diff_with_all_headers(self):
        """Testing Git diff parsing and preserving all headers"""
        preamble = (
            b'From 38d8fa94a9aa0c5b27943bec31d94e880165f1e0 Mon Sep '
            b'17 00:00:00 2001\n'
            b'From: Example Joe <joe@example.com>\n'
            b'Date: Thu, 5 Apr 2012 00:41:12 -0700\n'
            b'Subject: [PATCH 1/1] Sample patch.\n'
            b'\n'
            b'This is a test summary.\n'
            b'\n'
            b'With a description.\n'
            b'---\n'
            b' foo/bar |   2 -+n'
            b' README  |   2 -+n'
            b' 2 files changed, 2 insertions(+), 2 deletions(-)\n'
            b'\n'
        )
        diff1 = (
            b'diff --git a/foo/bar b/foo/bar2\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar2\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
        )
        diff2 = (
            b'diff --git a/README b/README\n'
            b'index 712544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/README\n'
            b'+++ b/README\n'
            b'@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
            b'-\n'
            b'1.7.1\n'
        )
        diff = preamble + diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo/bar',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo/bar2',
            modified_file_details=b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            delete_count=1,
            data=preamble + diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'README',
            orig_file_details=b'712544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'README',
            modified_file_details=b'f88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100644',
            new_unix_mode='100644',
            insert_count=1,
            delete_count=2,
            data=diff2)

    def test_parse_diff_revision(self):
        """Testing Git revision number parsing"""
        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'doc/readme',
                                          revision=b'bf544ea'),
            (b'doc/readme', b'bf544ea'))
        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'/dev/null',
                                          revision=b'bf544ea'),
            (b'/dev/null', PRE_CREATION))
        self.assertEqual(
            self.tool.parse_diff_revision(filename=b'/dev/null',
                                          revision=b'0000000'),
            (b'/dev/null', PRE_CREATION))

    def test_parse_diff_with_copy_and_rename_same_file(self):
        """Testing Git diff parsing with copy and rename of same file"""
        diff1 = (
            b'diff --git a/foo/bar b/foo/bar2\n'
            b'similarity index 100%\n'
            b'copy from foo/bar\n'
            b'copy to foo/bar2\n'
        )
        diff2 = (
            b'diff --git a/foo/bar b/foo/bar3\n'
            b'similarity index 92%\n'
            b'rename from foo/bar\n'
            b'rename to foo/bar3\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1 100644\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar3\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
        )
        diff = diff1 + diff2

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 2)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo/bar',
            orig_file_details=b'',
            modified_filename=b'foo/bar2',
            modified_file_details=b'',
            copied=True,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'foo/bar',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo/bar3',
            modified_file_details=b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100644',
            new_unix_mode='100644',
            moved=True,
            insert_count=1,
            delete_count=1,
            data=diff2)

    def test_parse_diff_with_mode_change_and_rename(self):
        """Testing Git diff parsing with mode change and rename"""
        diff = (
            b'diff --git a/foo/bar b/foo/bar2\n'
            b'old mode 100755\n'
            b'new mode 100644\n'
            b'similarity index 99%\n'
            b'rename from foo/bar\n'
            b'rename to foo/bar2\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1\n'
            b'--- a/foo/bar\n'
            b'+++ b/foo/bar2\n'
            b'@@ -1,1 +1,1 @@\n'
            b'-blah blah\n'
            b'+blah\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo/bar',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo/bar2',
            modified_file_details=b'e88b7f15c03d141d0bb38c8e49bb6c411ebfe1f1',
            old_unix_mode='100755',
            new_unix_mode='100644',
            moved=True,
            insert_count=1,
            delete_count=1,
            data=diff)

    def test_diff_git_line_without_a_b(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes
        """
        diff = (
            b'diff --git foo foo\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff)

    def test_diff_git_line_without_a_b_quotes(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with quotes
        """
        diff = (
            b'diff --git "foo" "foo"\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff)

    def test_diff_git_line_without_a_b_and_spaces(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces
        """
        diff = (
            b'diff --git foo bar1 foo bar1\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo bar1',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo bar1',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff)

    def test_diff_git_line_without_a_b_and_spaces_quotes(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with space and quotes
        """
        diff = (
            b'diff --git "foo bar1" "foo bar1"\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo bar1',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo bar1',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff)

    def test_diff_git_line_without_a_b_and_spaces_changed(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces, with filename changes
        """
        diff = (b'diff --git foo bar1 foo bar2\n'
                b'deleted file mode 100644\n'
                b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
                b'0000000000000000000000000000000000000000\n')

        with self.assertRaises(DiffParserError) as cm:
            self.tool.get_parser(diff).parse()

        self.assertTrue(str(cm.exception).startswith(
            'Unable to parse the "diff --git" line'))

    def test_diff_git_line_without_a_b_and_spaces_quotes_changed(self):
        """Testing parsing Git diff with deleted file without a/ and
        b/ filename prefixes and with spaces and quotes, with filename
        changes
        """
        diff1 = (
            b'diff --git "foo bar1" "foo bar2"\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )
        diff2 = (
            b'diff --git "foo bar1" foo\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )
        diff3 = (
            b'diff --git foo "foo bar1"\n'
            b'deleted file mode 100644\n'
            b'index 612544e4343bf04967eb5ea80257f6c64d6f42c7..'
            b'0000000000000000000000000000000000000000\n'
        )
        diff = diff1 + diff2 + diff3

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 3)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'foo bar1',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo bar2',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff1)

        self.assert_parsed_diff_file(
            parsed_files[1],
            orig_filename=b'foo bar1',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff2)

        self.assert_parsed_diff_file(
            parsed_files[2],
            orig_filename=b'foo',
            orig_file_details=b'612544e4343bf04967eb5ea80257f6c64d6f42c7',
            modified_filename=b'foo bar1',
            modified_file_details=b'0000000000000000000000000000000000000000',
            old_unix_mode='100644',
            deleted=True,
            data=diff3)

    def test_diff_git_symlink_added(self):
        """Testing parsing Git diff with symlink added"""
        diff = (
            b'diff --git a/link b/link\n'
            b'new file mode 120000\n'
            b'index 0000000..100b938\n'
            b'--- /dev/null\n'
            b'+++ b/link\n'
            b'@@ -0,0 +1 @@\n'
            b'+README\n'
            b'\\ No newline at end of file\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'link',
            orig_file_details=PRE_CREATION,
            modified_filename=b'link',
            modified_file_details=b'100b938',
            new_unix_mode='120000',
            is_symlink=True,
            new_symlink_target=b'README',
            insert_count=1,
            data=diff)

    def test_diff_git_symlink_changed(self):
        """Testing parsing Git diff with symlink changed"""
        diff = (
            b'diff --git a/link b/link\n'
            b'index 100b937..100b938 120000\n'
            b'--- a/link\n'
            b'+++ b/link\n'
            b'@@ -1 +1 @@\n'
            b'-README\n'
            b'\\ No newline at end of file\n'
            b'+README.md\n'
            b'\\ No newline at end of file\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'link',
            orig_file_details=b'100b937',
            modified_filename=b'link',
            modified_file_details=b'100b938',
            old_unix_mode='120000',
            new_unix_mode='120000',
            is_symlink=True,
            old_symlink_target=b'README',
            new_symlink_target=b'README.md',
            insert_count=1,
            delete_count=1,
            data=diff)

    def test_diff_git_symlink_removed(self):
        """Testing parsing Git diff with symlink removed"""
        diff = (
            b'diff --git a/link b/link\n'
            b'deleted file mode 120000\n'
            b'index 100b938..0000000\n'
            b'--- a/link\n'
            b'+++ /dev/null\n'
            b'@@ -1 +0,0 @@\n'
            b'-README.txt\n'
            b'\\ No newline at end of file\n'
        )

        parsed_files = self.tool.get_parser(diff).parse()
        self.assertEqual(len(parsed_files), 1)

        self.assert_parsed_diff_file(
            parsed_files[0],
            orig_filename=b'link',
            orig_file_details=b'100b938',
            modified_filename=b'link',
            modified_file_details=b'0000000',
            old_unix_mode='120000',
            is_symlink=True,
            old_symlink_target=b'README.txt',
            deleted=True,
            delete_count=1,
            data=diff)

    def test_file_exists(self):
        """Testing GitTool.file_exists"""
        tool = self.tool

        self.assertTrue(tool.file_exists('readme', 'e965047'))
        self.assertTrue(tool.file_exists('readme', 'd6613f5'))

        self.assertFalse(tool.file_exists('readme', PRE_CREATION))
        self.assertFalse(tool.file_exists('readme', 'fffffff'))
        self.assertFalse(tool.file_exists('readme2', 'fffffff'))

        # These sha's are valid, but commit and tree objects, not blobs.
        self.assertFalse(tool.file_exists('readme', 'a62df6c'))
        self.assertFalse(tool.file_exists('readme2', 'ccffbb4'))

    def test_get_file(self):
        """Testing GitTool.get_file"""
        tool = self.tool

        content = tool.get_file('readme', PRE_CREATION)
        self.assertIsInstance(content, bytes)
        self.assertEqual(content, b'')

        content = tool.get_file('readme', 'e965047')
        self.assertIsInstance(content, bytes)
        self.assertEqual(content, b'Hello\n')

        content = tool.get_file('readme', 'd6613f5')
        self.assertIsInstance(content, bytes)
        self.assertEqual(content, b'Hello there\n')

        content = tool.get_file('readme')
        self.assertIsInstance(content, bytes)
        self.assertEqual(content, b'Hello there\n')

        with self.assertRaises(SCMError):
            tool.get_file('')

        with self.assertRaises(FileNotFoundError):
            tool.get_file('', '0000000')

        with self.assertRaises(FileNotFoundError):
            tool.get_file('hello', '0000000')

        with self.assertRaises(FileNotFoundError):
            tool.get_file('readme', '0000000')

    def test_parse_diff_revision_with_remote_and_short_SHA1_error(self):
        """Testing GitTool.parse_diff_revision with remote files and short
        SHA1 error
        """
        with self.assertRaises(ShortSHA1Error):
            self.remote_tool.parse_diff_revision(filename=b'README',
                                                 revision=b'd7e96b3')

    def test_get_file_with_remote_and_short_SHA1_error(self):
        """Testing GitTool.get_file with remote files and short SHA1 error"""
        with self.assertRaises(ShortSHA1Error):
            self.remote_tool.get_file('README', 'd7e96b3')

    def test_valid_repository_https_username(self):
        """Testing GitClient.is_valid_repository with an HTTPS URL and external
        credentials
        """
        client = GitClient('https://example.com/test.git',
                           username='username',
                           password='pass/word')

        self.spy_on(client._run_git)
        client.is_valid_repository()

        self.assertEqual(client._run_git.calls[0].args[0],
                         ['ls-remote',
                          'https://username:pass%2Fword@example.com/test.git',
                          'HEAD'])

    def test_raw_file_url_error(self):
        """Testing Repository.get_file re-fetches when raw file URL changes"""
        repository = self.remote_repository

        self.spy_on(repository._get_file_uncached,
                    op=kgb.SpyOpReturn(b'first'))

        self.assertEqual(repository.get_file(path='PATH',
                                             revision='d7e96b3'),
                         b'first')

        # Ensure output of fake result matches.
        repository._get_file_uncached.unspy()
        self.spy_on(repository._get_file_uncached,
                    op=kgb.SpyOpReturn(b'second'))

        # Grab from cache when no changes and change fake result to confirm
        # it is not called.
        self.assertEqual(repository.get_file(path='PATH',
                                             revision='d7e96b3'),
                         b'first')

        # When raw_file_url changed, do not grab from cache and ensure output
        # equals second fake value.
        repository.raw_file_url = \
            'http://github.com/api/v2/yaml/blob/show/reviewboard/<revision>'

        self.assertEqual(repository.get_file(path='PATH',
                                             revision='d7e96b3'),
                         b'second')

    def test_get_file_exists_caching_with_raw_url(self):
        """Testing Repository.get_file_exists properly checks file existence in
        repository or cache when raw file URL changes
        """
        repository = self.remote_repository

        self.spy_on(repository._get_file_exists_uncached,
                    op=kgb.SpyOpReturn(True))

        # Use spy to put key into cache
        self.assertTrue(repository.get_file_exists(path='PATH',
                                                   revision='d7e96b3'))

        # Remove spy to ensure key is still in cache without needing spy
        repository._get_file_exists_uncached.unspy()
        self.assertTrue(repository.get_file_exists(path='PATH',
                                                   revision='d7e96b3'))

        # Does not exist when raw_file_url changed because it is not cached.
        repository.raw_file_url = \
            'http://github.com/api/v2/yaml/blob/show/reviewboard/<revision>'

        self.assertFalse(repository.get_file_exists(path='PATH',
                                                    revision='d7e96b3'))

    def test_normalize_patch_with_git_diff_new_symlink(self):
        """Testing GitTool.normalize_patch with new symlink"""
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git /dev/null b/test\n'
                    b'new file mode 120000\n'
                    b'--- /dev/null\n'
                    b'+++ b/test\n'
                    b'@@ -0,0 +1,1 @@\n'
                    b'+target_file\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision=PRE_CREATION),
            (
                b'diff --git /dev/null b/test\n'
                b'new file mode 100000\n'
                b'--- /dev/null\n'
                b'+++ b/test\n'
                b'@@ -0,0 +1,1 @@\n'
                b'+target_file\n'
                b'\\ No newline at end of file'
            ))

    def test_normalize_patch_with_modified_symlink(self):
        """Testing GitTool.normalize_patch with modified symlink"""
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git a/test b/test\n'
                    b'index abc1234..def4567 120000\n'
                    b'--- a/test\n'
                    b'+++ b/test\n'
                    b'@@ -1,1 +1,1 @@\n'
                    b'-old_target\n'
                    b'\\ No newline at end of file'
                    b'+new_target\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision='abc1234'),
            (
                b'diff --git a/test b/test\n'
                b'index abc1234..def4567 100000\n'
                b'--- a/test\n'
                b'+++ b/test\n'
                b'@@ -1,1 +1,1 @@\n'
                b'-old_target\n'
                b'\\ No newline at end of file'
                b'+new_target\n'
                b'\\ No newline at end of file'
            ))

    def test_normalize_patch_with_deleted_symlink(self):
        """Testing HgTool.normalize_patch with deleted symlink"""
        self.assertEqual(
            self.tool.normalize_patch(
                patch=(
                    b'diff --git a/test b/test\n'
                    b'deleted file mode 120000\n'
                    b'index abc1234..0000000\n'
                    b'--- a/test\n'
                    b'+++ /dev/null\n'
                    b'@@ -1,1 +0,0 @@\n'
                    b'-old_target\n'
                    b'\\ No newline at end of file'
                ),
                filename='test',
                revision='abc1234'),
            (
                b'diff --git a/test b/test\n'
                b'deleted file mode 100000\n'
                b'index abc1234..0000000\n'
                b'--- a/test\n'
                b'+++ /dev/null\n'
                b'@@ -1,1 +0,0 @@\n'
                b'-old_target\n'
                b'\\ No newline at end of file'
            ))


class GitAuthFormTests(TestCase):
    """Unit tests for GitTool's authentication form."""

    def test_fields(self):
        """Testing GitTool authentication form fields"""
        form = GitTool.create_auth_form()

        self.assertEqual(list(form.fields), ['username', 'password'])
        self.assertEqual(form['username'].help_text, '')
        self.assertEqual(form['username'].label, 'Username')
        self.assertEqual(form['password'].help_text, '')
        self.assertEqual(form['password'].label, 'Password')

    @add_fixtures(['test_scmtools'])
    def test_load(self):
        """Testing GitTool authentication form load"""
        repository = self.create_repository(
            tool_name='Git',
            username='test-user',
            password='test-pass')

        form = GitTool.create_auth_form(repository=repository)
        form.load()

        self.assertEqual(form['username'].value(), 'test-user')
        self.assertEqual(form['password'].value(), 'test-pass')

    @add_fixtures(['test_scmtools'])
    def test_save(self):
        """Testing GitTool authentication form save"""
        repository = self.create_repository(tool_name='Git')

        form = GitTool.create_auth_form(
            repository=repository,
            data={
                'username': 'test-user',
                'password': 'test-pass',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.username, 'test-user')
        self.assertEqual(repository.password, 'test-pass')


class GitRepositoryFormTests(TestCase):
    """Unit tests for GitTool's repository form."""

    def test_fields(self) -> None:
        """Testing GitTool repository form fields"""
        form = GitTool.create_repository_form()

        self.assertEqual(list(form.fields),
                         ['path', 'mirror_path', 'raw_file_url'])
        self.assertEqual(form['path'].help_text,
                         'For local Git repositories, this should be the path '
                         'to a .git directory that Review Board can read '
                         'from. For remote Git repositories, it should be '
                         'the clone URL.')
        self.assertEqual(form['path'].label, 'Path')
        self.assertEqual(form['mirror_path'].help_text, '')
        self.assertEqual(form['mirror_path'].label, 'Mirror Path')
        self.assertEqual(form['raw_file_url'].label, 'Raw File URL Mask')
        self.assertEqual(form['raw_file_url'].help_text,
                         "A URL mask used to check out a particular revision "
                         "of a file using HTTP. This is needed for "
                         "repository types that can't access remote files "
                         "natively. Use <tt>&lt;revision&gt;</tt> and "
                         "<tt>&lt;filename&gt;</tt> in the URL in place of "
                         "the revision and filename parts of the path.")

    @add_fixtures(['test_scmtools'])
    def test_load(self) -> None:
        """Testing GitTool repository form load"""
        repository = self.create_repository(
            tool_name='Git',
            path='https://github.com/reviewboard/reviewboard',
            mirror_path='git@github.com:reviewboard/reviewboard.git',
            raw_file_url='http://git.example.com/raw/<revision>')

        form = GitTool.create_repository_form(repository=repository)
        form.load()

        self.assertEqual(form['path'].value(),
                         'https://github.com/reviewboard/reviewboard')
        self.assertEqual(form['mirror_path'].value(),
                         'git@github.com:reviewboard/reviewboard.git')
        self.assertEqual(form['raw_file_url'].value(),
                         'http://git.example.com/raw/<revision>')

    @add_fixtures(['test_scmtools'])
    def test_save(self) -> None:
        """Testing GitTool repository form save"""
        repository = self.create_repository(tool_name='Git')

        form = GitTool.create_repository_form(
            repository=repository,
            data={
                'path': 'https://github.com/reviewboard/reviewboard',
                'mirror_path': 'git@github.com:reviewboard/reviewboard.git',
                'raw_file_url': 'http://git.example.com/raw/<revision>',
            })
        self.assertTrue(form.is_valid())
        form.save()

        self.assertEqual(repository.path,
                         'https://github.com/reviewboard/reviewboard')
        self.assertEqual(repository.mirror_path,
                         'git@github.com:reviewboard/reviewboard.git')
        self.assertEqual(repository.raw_file_url,
                         'http://git.example.com/raw/<revision>')

    @add_fixtures(['test_scmtools'])
    def test_save_with_local_path(self) -> None:
        """Testing GitTool repository form save with a local .git path"""
        repository = self.create_repository(tool_name='Git')

        form = GitTool.create_repository_form(
            repository=repository,
            data={
                'path': '/opt/repositories/repo1/.git',
            })
        form.full_clean()

        self.assertTrue(form.is_valid())

        form.save()
        self.assertEqual(repository.path, '/opt/repositories/repo1/.git')

    @add_fixtures(['test_scmtools'])
    def test_save_with_file_url(self) -> None:
        """Testing GitTool repository form save with a file:// URL"""
        repository = self.create_repository(tool_name='Git')

        form = GitTool.create_repository_form(
            repository=repository,
            data={
                'path': 'file:///opt/repositories/repo1/.git',
            })
        form.full_clean()

        self.assertTrue(form.is_valid())

        form.save()
        self.assertEqual(repository.path,
                         'file:///opt/repositories/repo1/.git')

    @add_fixtures(['test_scmtools'])
    def test_save_with_no_raw_file(self) -> None:
        """Testing GitTool repository form save with a remote repo and no raw
        file mask
        """
        repository = self.create_repository(tool_name='Git')

        form = GitTool.create_repository_form(
            repository=repository,
            data={
                'path': 'https://github.com/reviewboard/reviewboard',
                'mirror_path': 'git@github.com:reviewboard/reviewboard.git',
            })
        form.full_clean()

        self.assertFalse(form.is_valid())

        manual_url = get_manual_url() + 'admin/configuration/repositories/git/'
        self.assertEqual(form.errors['path'], [
            'Remote Git repositories cannot be accessed without a Raw File '
            'URL Mask. See the <a href="%s">documentation</a> for more '
            'details.'
            % manual_url,
        ])



"""Add Profile.should_send_own_updates.

Version Added:
    2.0.9
"""

from django_evolution.mutations import AddField
from django.db import models


MUTATIONS = [
    AddField('Profile', 'should_send_own_updates', models.BooleanField,
             initial=True)
]



"""A hook for adding a new widget to the administration dashboard."""

from __future__ import annotations

from djblets.extensions.hooks import BaseRegistryHook, ExtensionHookPoint

from reviewboard.admin.widgets import admin_widgets_registry


class AdminWidgetHook(BaseRegistryHook, metaclass=ExtensionHookPoint):
    """A hook for adding a new widget to the administration dashboard.

    Version Changed:
        4.0:
        Widget classes should now subclass
        :py:class:`~reviewboard.admin.widgets.AdminBaseWidget` instead of
        :py:class:`~reviewboard.admin.widgets.Widget`. Note that this will
        require a full rewrite of the widget.

        The ``primary`` argument is no longer supported when instantiating
        the hook, and will be ignored. Callers should remove it.

    Version Changed:
        5.0:
        Support for legacy widgets and arguments has been removed.
    """

    registry = admin_widgets_registry



"""Test cases for the hosting service client support."""

from __future__ import annotations

from typing import TYPE_CHECKING

from kgb import SpyAgency

from reviewboard.deprecation import RemovedInReviewBoard90Warning
from reviewboard.hostingsvcs.base import (HostingServiceClient,
                                          HostingServiceHTTPRequest,
                                          HostingServiceHTTPResponse)
from reviewboard.hostingsvcs.models import HostingServiceAccount
from reviewboard.testing.hosting_services import TestService
from reviewboard.testing.testcase import TestCase

if TYPE_CHECKING:
    from djblets.testing.testcases import ExpectedWarning


class _DummyHTTPRequest(HostingServiceHTTPRequest):
    def open(self) -> HostingServiceHTTPResponse:
        method = self.method

        if method in {'DELETE', 'HEAD'}:
            data = None
        else:
            data = b'{"key": "test response"}'

        if method == 'DELETE':
            status_code = 204
        elif method == 'POST':
            status_code = 201
        else:
            status_code = 200

        return HostingServiceHTTPResponse(
            request=self,
            url=self.url,
            data=data,
            headers={
                'Test-header': 'Value',
            },
            status_code=status_code)


class HostingServiceHTTPRequestTests(TestCase):
    """Unit tests for HostingServiceHTTPRequest."""

    def test_init_with_query(self) -> None:
        """Testing HostingServiceHTTPRequest construction with query="""
        request = HostingServiceHTTPRequest(
            url='http://example.com?z=1&z=2&baz=true',
            query={
                'foo': 'bar',
                'a': 10,
                'list': ['a', 'b', 'c'],
            })

        self.assertEqual(
            request.url,
            'http://example.com?a=10&baz=true&foo=bar&list=a&list=b&list=c'
            '&z=1&z=2')

    def test_init_with_body_not_bytes(self) -> None:
        """Testing HostingServiceHTTPRequest construction with non-bytes body
        """
        account = HostingServiceAccount()
        service = TestService(account)

        expected_message = (
            f'Received non-bytes body for the HTTP request for '
            f'{TestService!r}. This is likely an implementation problem. '
            f'Please make sure only byte strings are sent for the request '
            f'body.'
        )

        with self.assertRaisesMessage(TypeError, expected_message):
            HostingServiceHTTPRequest(
                url='http://example.com?z=1&z=2&baz=true',
                method='POST',
                body=123,  # type:ignore
                hosting_service=service)

    def test_init_with_header_key_not_unicode(self) -> None:
        """Testing HostingServiceHTTPRequest construction with non-Unicode
        header key
        """
        account = HostingServiceAccount()
        service = TestService(account)

        expected_message = (
            f"Received non-Unicode header name b'My-Header' (value='abc') for "
            f"the HTTP request for {TestService!r}. This is likely an "
            f"implementation problem. Please make sure only Unicode strings "
            f"are sent in request headers."
        )

        with self.assertRaisesMessage(TypeError, expected_message):
            HostingServiceHTTPRequest(
                url='http://example.com?z=1&z=2&baz=true',
                method='POST',
                headers={  # type:ignore
                    b'My-Header': 'abc',
                },
                hosting_service=service)

    def test_init_with_header_value_not_unicode(self) -> None:
        """Testing HostingServiceHTTPRequest construction with non-Unicode
        header value
        """
        account = HostingServiceAccount()
        service = TestService(account)

        expected_message = (
            f"Received non-Unicode header value for 'My-Header' "
            f"(value=b'abc') for the HTTP request for {TestService!r}. This "
            f"is likely an implementation problem. Please make sure only "
            f"Unicode strings are sent in request headers."
        )

        with self.assertRaisesMessage(TypeError, expected_message):
            HostingServiceHTTPRequest(
                url='http://example.com?z=1&z=2&baz=true',
                method='POST',
                headers={  # type:ignore
                    'My-Header': b'abc',
                },
                hosting_service=service)

    def test_add_basic_auth(self) -> None:
        """Testing HostingServiceHTTPRequest.add_basic_auth"""
        request = HostingServiceHTTPRequest('http://example.com')
        request.add_basic_auth(b'username', b'password')

        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
            })

    def test_get_header(self) -> None:
        """Testing HostingServiceHTTPRequest.get_header"""
        request = HostingServiceHTTPRequest(
            'http://example.com',
            headers={
                'Authorization': 'Basic abc123',
                'Content-Length': '123',
            })

        self.assertEqual(request.get_header('Authorization'), 'Basic abc123')
        self.assertEqual(request.get_header('AUTHORIZATION'), 'Basic abc123')
        self.assertEqual(request.get_header('authorization'), 'Basic abc123')

        self.assertEqual(request.get_header('Content-Length'), '123')
        self.assertEqual(request.get_header('CONTENT-LENGTH'), '123')
        self.assertEqual(request.get_header('content-length'), '123')


class HostingServiceHTTPResponseTests(TestCase):
    """Unit tests for HostingServiceHTTPResponse."""

    def test_json(self) -> None:
        """Testing HostingServiceHTTPResponse.json"""
        request = HostingServiceHTTPRequest('http://example.com')
        response = HostingServiceHTTPResponse(request=request,
                                              url='http://example.com',
                                              data=b'{"a": 1, "b": 2}',
                                              headers={},
                                              status_code=200)
        self.assertEqual(
            response.json,
            {
                'a': 1,
                'b': 2,
            })

    def test_json_with_non_json_response(self) -> None:
        """Testing HostingServiceHTTPResponse.json with non-JSON response"""
        request = HostingServiceHTTPRequest('http://example.com')
        response = HostingServiceHTTPResponse(request=request,
                                              url='http://example.com',
                                              data=b'XXX',
                                              headers={},
                                              status_code=200)

        with self.assertRaises(ValueError):
            response.json

    def test_get_header(self) -> None:
        """Testing HostingServiceHTTPRequest.get_header"""
        request = HostingServiceHTTPRequest('http://example.com')
        response = HostingServiceHTTPResponse(
            request=request,
            url=request.url,
            status_code=200,
            data=b'',
            headers={
                'Authorization': 'Basic abc123',
                'Content-Length': '123',
            })

        self.assertEqual(response.get_header('Authorization'), 'Basic abc123')
        self.assertEqual(response.get_header('AUTHORIZATION'), 'Basic abc123')
        self.assertEqual(response.get_header('authorization'), 'Basic abc123')

        self.assertEqual(response.get_header('Content-Length'), '123')
        self.assertEqual(response.get_header('CONTENT-LENGTH'), '123')
        self.assertEqual(response.get_header('content-length'), '123')


class HostingServiceClientTests(SpyAgency, TestCase):
    """Unit tests for HostingServiceClient"""

    #: The hosting service client for the tests.
    #:
    #: Type:
    #:     reviewboard.hostingsvcs.base.client.HostingServiceClient
    client: HostingServiceClient

    def setUp(self) -> None:
        """Set up the test case."""
        super().setUp()

        account = HostingServiceAccount()
        service = TestService(account)

        self.client = HostingServiceClient(service)
        self.client.http_request_cls = _DummyHTTPRequest

    def tearDown(self) -> None:
        """Tear down the test case."""
        super().tearDown()

        self.client = None  # type: ignore

    def test_http_delete(self) -> None:
        """Testing HostingServiceClient.http_delete"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_delete(
            url='http://example.com',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.data, b'')
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.status_code, 204)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=None,
            headers={
                'Foo': 'bar',
            },
            credentials={
                'username': 'username',
                'password': 'password',
            })

        request = self.client.build_http_request.last_call.return_value
        self.assertIsNone(request.body)
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'DELETE')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_http_get(self) -> None:
        """Testing HostingServiceClient.http_get"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_get(
            url='http://example.com',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=None,
            headers={
                'Foo': 'bar',
            },
            method='GET',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertIsNone(request.body)
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'GET')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_http_head(self) -> None:
        """Testing HostingServiceClient.http_head"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_head(
            url='http://example.com',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.data, b'')
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=None,
            headers={
                'Foo': 'bar',
            },
            method='HEAD',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertIsNone(request.body)
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'HEAD')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_http_post_with_body_unicode(self) -> None:
        """Testing HostingServiceClient.http_post with body as Unicode"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_post(
            url='http://example.com',
            body='test body\U0001f60b'.encode('utf-8'),
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 201)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\xf0\x9f\x98\x8b',
            headers={
                'Content-Length': '13',
                'Foo': 'bar',
            },
            method='POST',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(request.body, b'test body\xf0\x9f\x98\x8b')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '13',
                'Foo': 'bar',
            })

    def test_http_post_with_body_bytes(self) -> None:
        """Testing HostingServiceClient.http_post with body as bytes"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_post(
            url='http://example.com',
            body=b'test body\x01\x02\x03',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 201)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\x01\x02\x03',
            headers={
                'Content-Length': '12',
                'Foo': 'bar',
            },
            method='POST',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(request.body, b'test body\x01\x02\x03')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '12',
                'Foo': 'bar',
            })

    def test_http_put_with_body_unicode(self) -> None:
        """Testing HostingServiceClient.http_put with body as Unicode"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_put(
            url='http://example.com',
            body='test body\U0001f60b',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\xf0\x9f\x98\x8b',
            headers={
                'Content-Length': '13',
                'Foo': 'bar',
            },
            method='PUT',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'PUT')
        self.assertEqual(request.body, b'test body\xf0\x9f\x98\x8b')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '13',
                'Foo': 'bar',
            })

    def test_http_put_with_body_bytes(self) -> None:
        """Testing HostingServiceClient.http_put with body as bytes"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_put(
            url='http://example.com',
            body=b'test body\x01\x02\x03',
            headers={
                'Foo': 'bar',
            },
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\x01\x02\x03',
            headers={
                'Content-Length': '12',
                'Foo': 'bar',
            },
            method='PUT',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'PUT')
        self.assertEqual(request.body, b'test body\x01\x02\x03')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '12',
                'Foo': 'bar',
            })

    def test_http_request(self) -> None:
        """Testing HostingServiceClient.http_request"""
        self.spy_on(self.client.build_http_request)

        response = self.client.http_request(
            url='http://example.com',
            body=b'test',
            headers={
                'Foo': 'bar',
            },
            method='BAZ',
            username='username',
            password='password')

        self.assertIsInstance(response, HostingServiceHTTPResponse)
        self.assertEqual(response.url, 'http://example.com')
        self.assertEqual(response.data, b'{"key": "test response"}')
        self.assertEqual(response.status_code, 200)
        self.assertIsInstance(response.headers, dict)
        self.assertEqual(
            response.headers,
            {
                'Test-header': 'Value',
            })

        # One for each item in the tuple, + 1 to detect the bounds.
        expected_warnings: list[ExpectedWarning] = [
            {
                'cls': RemovedInReviewBoard90Warning,
                'message': (
                    'Accessing HostingServiceHTTPResponse by index is '
                    'deprecated. Please use HostingServiceHTTPResponse.data '
                    'or HostingServiceHTTPResponse.headers instead. This will '
                    'be removed in Review Board 9.'
                ),
            }
            for _i in range(3)
        ]

        with self.assertWarnings(expected_warnings):
            data, headers = response

        self.assertEqual(data, response.data)
        self.assertEqual(headers, response.headers)

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test',
            headers={
                'Foo': 'bar',
            },
            method='BAZ',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'BAZ')
        self.assertEqual(request.body, b'test')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_build_http_request(self) -> None:
        """Testing HostingServiceClient.build_http_request"""
        request = self.client.build_http_request(
            url='http://example.com',
            body=b'test',
            method='POST',
            credentials={},
            headers={
                'Foo': 'bar',
            })

        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.body, b'test')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(
            request.headers,
            {
                'Foo': 'bar',
            })

    def test_build_http_request_with_basic_auth(self) -> None:
        """Testing HostingServiceClient.build_http_request with username and
        password
        """
        request = self.client.build_http_request(
            url='http://example.com',
            body=b'test',
            method='POST',
            headers={
                'Foo': 'bar',
            },
            credentials={
                'username': 'username',
                'password': 'password',
            })

        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.body, b'test')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_json_delete(self) -> None:
        """Testing HostingServiceClient.json_delete"""
        self.spy_on(self.client.build_http_request)

        message = (
            'HostingServiceClient.json_delete is deprecated. Please use '
            'HostingServiceClient.http_delete instead. This will be removed '
            'in Review Board 9.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning, message):
            rsp, headers = self.client.json_delete(
                url='http://example.com',
                headers={
                    'Foo': 'bar',
                },
                username='username',
                password='password')

        self.assertEqual(rsp, b'')
        self.assertIsInstance(headers, dict)
        self.assertEqual(
            headers,
            {
                'Test-header': 'Value',
            })

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=None,
            headers={
                'Foo': 'bar',
            },
            credentials={
                'username': 'username',
                'password': 'password',
            })

        request = self.client.build_http_request.last_call.return_value
        self.assertIsNone(request.body)
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'DELETE')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_json_get(self) -> None:
        """Testing HostingServiceClient.json_get"""
        self.spy_on(self.client.build_http_request)

        message = (
            'HostingServiceClient.json_get is deprecated. Please use '
            'HostingServiceClient.http_get instead. This will be removed '
            'in Review Board 9.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning, message):
            rsp, headers = self.client.json_get(
                url='http://example.com',
                headers={
                    'Foo': 'bar',
                },
                username='username',
                password='password')

        self.assertEqual(
            rsp,
            {
                'key': 'test response',
            })
        self.assertIsInstance(headers, dict)
        self.assertEqual(
            headers,
            {
                'Test-header': 'Value',
            })

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=None,
            headers={
                'Foo': 'bar',
            },
            method='GET',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertIsNone(request.body)
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'GET')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Foo': 'bar',
            })

    def test_json_post_with_body_unicode(self) -> None:
        """Testing HostingServiceClient.json_post with body as Unicode"""
        self.spy_on(self.client.build_http_request)

        message = (
            'HostingServiceClient.json_post is deprecated. Please use '
            'HostingServiceClient.http_post instead. This will be removed '
            'in Review Board 9.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning, message):
            rsp, headers = self.client.json_post(
                url='http://example.com',
                body='test body\U0001f60b',
                headers={
                    'Foo': 'bar',
                },
                username='username',
                password='password')

        self.assertEqual(
            rsp,
            {
                'key': 'test response',
            })
        self.assertIsInstance(headers, dict)
        self.assertEqual(
            headers,
            {
                'Test-header': 'Value',
            })

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\xf0\x9f\x98\x8b',
            headers={
                'Content-Length': '13',
                'Foo': 'bar',
            },
            method='POST',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(request.body, b'test body\xf0\x9f\x98\x8b')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '13',
                'Foo': 'bar',
            })

    def test_json_post_with_body_bytes(self) -> None:
        """Testing HostingServiceClient.json_post with body as bytes"""
        self.spy_on(self.client.build_http_request)

        message = (
            'HostingServiceClient.json_post is deprecated. Please use '
            'HostingServiceClient.http_post instead. This will be removed '
            'in Review Board 9.'
        )

        with self.assertWarns(RemovedInReviewBoard90Warning, message):
            rsp, headers = self.client.json_post(
                url='http://example.com',
                body=b'test body\x01\x02\x03',
                headers={
                    'Foo': 'bar',
                },
                username='username',
                password='password')

        self.assertEqual(
            rsp,
            {
                'key': 'test response',
            })
        self.assertIsInstance(headers, dict)
        self.assertEqual(
            headers,
            {
                'Test-header': 'Value',
            })

        self.assertSpyCalledWith(
            self.client.build_http_request,
            url='http://example.com',
            body=b'test body\x01\x02\x03',
            headers={
                'Content-Length': '12',
                'Foo': 'bar',
            },
            method='POST',
            username='username',
            password='password')

        request = self.client.build_http_request.last_call.return_value
        self.assertEqual(request.url, 'http://example.com')
        self.assertEqual(request.method, 'POST')
        self.assertEqual(request.body, b'test body\x01\x02\x03')
        self.assertIsInstance(request.headers, dict)
        self.assertEqual(
            request.headers,
            {
                'Authorization': 'Basic dXNlcm5hbWU6cGFzc3dvcmQ=',
                'Content-length': '12',
                'Foo': 'bar',
            })



"""File attachment mimetype registration and scoring."""

from __future__ import annotations

from collections import OrderedDict
import logging
import os
import subprocess
from typing import ClassVar, Final, Optional, TYPE_CHECKING

import docutils.core
import mimeparse
from django.contrib.staticfiles.storage import staticfiles_storage
from django.templatetags.static import static
from django.utils.html import format_html, format_html_join
from django.utils.encoding import force_str, smart_str
from django.utils.safestring import mark_safe
from django.utils.translation import gettext as _
from djblets.cache.backend import cache_memoize
from djblets.siteconfig.models import SiteConfiguration
from djblets.util.filesystem import is_exe_in_path
from djblets.util.templatetags.djblets_images import thumbnail
from pygments import highlight
from pygments.lexers import (ClassNotFound, guess_lexer_for_filename,
                             TextLexer)

from reviewboard.reviews.markdown_utils import render_markdown
from reviewboard.webapi.server_info import get_capabilities

if TYPE_CHECKING:
    from django.core.files import File


logger = logging.getLogger(__name__)


_registered_mimetype_handlers = []
_supported_mimetypes: list[tuple[str, str, dict[str, str]]] = []


DEFAULT_MIMETYPE = 'application/octet-stream'


def guess_mimetype(
    uploaded_file: File,
) -> str:
    """Guess the mimetype of an uploaded file.

    Uploaded files don't necessarily have valid mimetypes provided,
    so attempt to guess them when they're blank.

    This only works if :program:`file` is in the path. If it's not, or
    guessing fails, we fall back to a mimetype of
    :mimetype:`application/octet-stream`.

    Args:
        uploaded_file (django.core.files.File):
            The uploaded file object.

    Returns:
        str:
        The guessed mimetype.
    """
    # Set the default mimetype.
    mimetype = DEFAULT_MIMETYPE

    if not is_exe_in_path('file'):
        return mimetype

    try:
        # The browser didn't know what this was, so we'll need to do
        # some guess work. If we have 'file' available, use that to
        # figure it out.
        with subprocess.Popen(['file', '--mime-type', '-b', '-'],
                              stdout=subprocess.PIPE,
                              stderr=subprocess.PIPE,
                              stdin=subprocess.PIPE) as p:
            assert p.stdin is not None
            assert p.stdout is not None

            # Write the content from the file until file has enough data
            # to make a determination.
            for chunk in uploaded_file.chunks():
                try:
                    p.stdin.write(chunk)
                except IOError:
                    # `file` closed the stream. It no longer needs any
                    # more input, so we can stop now. We hopefully have
                    # an answer.
                    break

            try:
                p.stdin.close()
            except IOError:
                # This was closed by `file`.
                #
                # Note that we may not get this on all Python
                # environments. A closed pipe doesn't necessarily fail
                # when calling close() again.
                pass

            ret = p.wait()

            if ret == 0:
                result = p.stdout.read().strip().decode('utf-8')

                if result:
                    mimetype = result
    except Exception as e:
        logger.exception('Unexpected error when determining mimetype '
                         'using `file`: %s',
                         e)
    finally:
        # Reset the read position so we can properly save this.
        uploaded_file.seek(0)

    return mimetype


def get_uploaded_file_mimetype(uploaded_file):
    """Return the mimetype of a file that was uploaded.

    There are several things that can go wrong with browser-provided mimetypes.
    In one case (bug 3427), Firefox on Linux Mint was providing a mimetype that
    looked like ``text/text/application/pdf``, which is unparsable. IE also
    has a habit of setting any unknown file type to
    :mimetype:`application/octet-stream`, rather than just choosing not to
    provide a mimetype. In the case where what we get from the browser is
    obviously wrong, try to guess.

    Args:
        uploaded_file (django.core.files.File):
            The uploaded file object.

    Returns:
        unicode:
        The guessed mimetype.
    """
    if (uploaded_file.content_type and
        len(uploaded_file.content_type.split('/')) == 2 and
            uploaded_file.content_type != 'application/octet-stream'):
        mimetype = uploaded_file.content_type
    else:
        mimetype = guess_mimetype(uploaded_file)

    return mimetype


def register_mimetype_handler(handler):
    """Register a MimetypeHandler class.

    This will register a mimetype Handler used by Review Board to render
    thumbnails for the file attachments across different mimetypes.

    Args:
        handler (type):
            The mimetype handler to register. This must be a subclass of
            :py:class:`MimetypeHandler`.

    Raises:
        TypeError:
            The provided class is not of the correct type.
    """
    if not issubclass(handler, MimetypeHandler):
        raise TypeError('Only MimetypeHandler subclasses can be registered')

    _registered_mimetype_handlers.append(handler)


def unregister_mimetype_handler(handler):
    """Unregister a MimetypeHandler class.

    This will unregister a previously registered mimetype handler.

    Args:
        handler (type):
            The mimetype handler to unregister. This must be a subclass of
            :py:class:`MimetypeHandler`.

    Raises:
        TypeError:
            The provided class is not of the correct type.

        ValueError:
            The mimetype handler was not previously registered.
    """
    if not issubclass(handler, MimetypeHandler):
        raise TypeError('Only MimetypeHandler subclasses can be unregistered')

    try:
        _registered_mimetype_handlers.remove(handler)
    except ValueError:
        logger.error('Failed to unregister missing mimetype handler %r',
                     handler)
        raise ValueError('This mimetype handler was not previously registered')


def score_match(pattern, test):
    """Return a score for how well the pattern matches the mimetype.

    This is an ordered list of precedence (``_`` indicates non-match):

    ======================= ==========
    Format                  Precedence
    ======================= ==========
    ``Type/Vendor+Subtype`` 2.0
    ``Type/Subtype``        1.9
    ``Type/*``              1.8
    ``*/Vendor+Subtype``    1.7
    ``*/_     +Subtype``    1.6
    ``*/*``                 1.5
    ``_``                   0
    ======================= ==========

    Args:
        pattern (tuple):
            A parsed mimetype pattern to score. This is a 3-tuple of the type,
            subtype, and parameters as returned by
            :py:func:`mimeparse.parse_mime_type`. This may include ``*``
            wildcards.

        test (tuple):
            A parsed mimetype to match against the pattern. This is a 3-tuple
            of the type, subtype, and parameters as returned by
            :py:func:`mimeparse.parse_mime_type`.

    Returns:
        float:
        The resulting score for the match.
    """
    def split_type(mimetype):
        """Split a mimetype into type/vendor/subtype.

        Args:
            mimetype (tuple):
                A parsed mimetype to match against the pattern. This is a
                3-tuple of the type, subtype, and parameters as returned by
                :py:func:`mimeparse.parse_mime_type`.

        Returns:
            tuple:
            A 3-tuple containing the type and split vendor/subtype.
        """
        subtype = mimetype[1].split('+', 1)

        if len(subtype) > 1:
            vendor = subtype[0]
            subtype = subtype[1]
        else:
            vendor = '*'
            subtype = subtype[0]

        return mimetype[0], subtype, vendor

    EXACT_TYPE = 1.0
    ANY_TYPE = 0.7
    EXACT_SUBTYPE = 0.9
    ANY_SUBTYPE = 0.8
    VND_SUBTYPE = 0.1

    pattern_type, pattern_subtype, pattern_vendor = split_type(pattern)
    test_type, test_subtype, test_vendor = split_type(test)

    score = 0

    if pattern_type == test_type:
        score += EXACT_TYPE
    elif pattern_type == '*':
        score += ANY_TYPE
    else:
        return 0

    if pattern_subtype == test_subtype:
        score += EXACT_SUBTYPE
    elif pattern_subtype == '*':
        score += ANY_SUBTYPE
    else:
        return 0

    if pattern_vendor != '*' and pattern_vendor == test_vendor:
        score += VND_SUBTYPE

    return score


def is_mimetype_supported(
    mimetype: str,
) -> bool:
    """Return whether the mimetype is supported by any review UI.

    Version Added:
        8.0

    Args:
        mimetype (str):
            The mimetype to check.

    Returns:
        bool:
        Whether the mimetype is supported by any review UI.
    """
    def _is_mimetype_supported(
            mimetype: str,
    ) -> bool:
        global _supported_mimetypes
        parsed = mimeparse.parse_mime_type(mimetype)

        if not _supported_mimetypes:
            capabilities = get_capabilities()
            _supported_mimetypes = [
                mimeparse.parse_mime_type(m)
                for m in capabilities['review_uis']['supported_mimetypes']
            ]

        supported = any(score_match(pattern, parsed)
                        for pattern in _supported_mimetypes)

        return supported

    return cache_memoize(f'review-uis:supported-mimetypes:{mimetype}',
                         lambda: _is_mimetype_supported(mimetype))


class MimetypeHandler(object):
    """Handles mimetype-specific properties.

    This class also acts as a generic handler for mimetypes not matched
    explicitly by any handler. Note that this is not the same as ``*/*``.

    Attributes:
        attachment (reviewboard.attachments.models.FileAttachment):
            The file attachment being handled.

        mimetype (unicode):
            The mimetype for the file attachment.
    """

    MIMETYPES_DIR: Final[str] = 'rb/images/mimetypes'

    #: All scaling factors for multi-DPI thumbnail images.
    #:
    #: This is only used for mimetype handlers that support thumbnail images.
    #:
    #: Version Added:
    #:     7.0
    THUMBNAIL_IMAGE_SCALES: Final[tuple[int, ...]] = (1, 2, 3)

    #: The base size for thumbnail images.
    #:
    #: This is only used for mimetype handlers that support thumbnail images.
    #:
    #: Version Added:
    #:     7.0
    BASE_THUMBNAIL_IMAGE_WIDTH: Final[int] = 300

    #: A list of mimetypes supported by this handler.
    supported_mimetypes: ClassVar[list[str]] = []

    #: Whether HD thumbnails are provided by this handler.
    #:
    #: Subclasses (especially in extensions) can use this to introspect what
    #: size thumbnails they should generate.
    use_hd_thumbnails: ClassVar[bool] = True

    def __init__(self, attachment, mimetype):
        """Initialize the handler.

        Args:
            attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment being handled.

            mimetype (unicode):
                The mimetype for the file attachment.
        """
        self.attachment = attachment
        self.mimetype = mimetype
        self.storage = staticfiles_storage

    @classmethod
    def get_best_handler(cls, mimetype):
        """Return the handler and score that that best fit the mimetype.

        Args:
            mimetype (tuple):
                A parsed mimetype to find the best handler for. This is a
                3-tuple of the type, subtype, and parameters as returned by
                :py:func:`mimeparse.parse_mime_type`.

        Returns:
            tuple:
            A tuple of ``(best_score, mimetype_handler)``. If no handler
            was found, this will be ``(0, None)``.
        """
        best_score, best_fit = (0, None)

        for mimetype_handler in _registered_mimetype_handlers:
            for mt in mimetype_handler.supported_mimetypes:
                try:
                    score = score_match(mimeparse.parse_mime_type(mt),
                                        mimetype)

                    if score > best_score:
                        best_score, best_fit = (score, mimetype_handler)
                except ValueError:
                    continue

        return (best_score, best_fit)

    @classmethod
    def for_type(cls, attachment):
        """Return the handler that is the best fit for provided mimetype.

        Args:
            attachment (reviewboard.attachments.models.FileAttachment):
                The file attachment to find the best handler for.

        Returns:
            MimetypeHandler:
            The best mimetype handler for the attachment, or ``None`` if
            one could not be found.
        """
        if not attachment.mimetype:
            return None

        try:
            mimetype = mimeparse.parse_mime_type(attachment.mimetype)
        except Exception:
            logger.warning('Unable to parse MIME type "%s" for %s',
                           attachment.mimetype, attachment)
            mimetype = ('application', 'octet-stream', {})

        # Override the mimetype if mimeparse is known to misinterpret this
        # type of file as `octet-stream`
        extension = os.path.splitext(attachment.filename)[1]

        if extension in MIMETYPE_EXTENSIONS:
            mimetype = MIMETYPE_EXTENSIONS[extension]

        score, handler = cls.get_best_handler(mimetype)

        if handler:
            try:
                return handler(attachment, mimetype)
            except Exception as e:
                logger.error('Unable to load Mimetype Handler for %s: %s',
                             attachment, e)

        return MimetypeHandler(attachment, mimetype)

    def get_icon_url(self):
        """Return the appropriate icon URL for this mimetype.

        Returns:
            unicode:
            The URL to an icon representing this mimetype.
        """
        mimetype_string = self.mimetype[0] + '/' + self.mimetype[1]

        if mimetype_string in MIMETYPE_ICON_ALIASES:
            path = self._get_mimetype_file(
                MIMETYPE_ICON_ALIASES[mimetype_string])
        else:
            path = self._get_mimetype_file(self.mimetype[0] + '-' +
                                           self.mimetype[1])
            if not self.storage.exists(path):
                path = self._get_mimetype_file(self.mimetype[0] + '-x-generic')

                if not self.storage.exists(path):
                    # We'll just use this as our fallback.
                    path = self._get_mimetype_file('text-x-generic')

        return static(path)

    def get_thumbnail(self):
        """Return HTML that represents a preview of the attachment.

        Subclasses can override this to provide a suitable thumbnail. The
        outer element of the thumbnail should have a ``file-thumbnail`` CSS
        class.

        By default, this returns an empty thumbnail.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the thumbnail for the associated attachment.
        """
        return mark_safe('<pre class="file-thumbnail"></pre>')

    def set_thumbnail(self, data):
        """Set the thumbnail data for this attachment.

        This should be implemented by subclasses if they need the thumbnail to
        be generated client-side.

        Args:
            data (bytes):
                The contents of the thumbnail data.
        """
        raise NotImplementedError

    def get_raw_thumbnail_image_url(
        self,
        *,
        width: Optional[int] = None,
        height: Optional[int] = None,
    ) -> Optional[str]:
        """Return the URL to a thumbnail of a given size.

        For mimetype handlers that support image-based thumbnails, this will
        compute a thumbnail matching the given size requirements, and then
        return the URL to that thumbnail in file storage.

        The caller must provide a width or a height at minimum, and may provide
        both. If providing only a width or a height, the thumbnail size is
        expected to be constrained only to that dimension.

        Subclasses must override :py:meth:`generate_thumbnail_image` to
        provide the logic for creating a thumbnail.

        Version Added:
            7.0

        Args:
            width (int, optional):
                The width to constrain the thumbnail image to.

            height (int, optional):
                The height to constrain the thumbnail image to.

        Raises:
            NotImplementedError:
                The mimetype handler does not support thumbnail images.

            ValueError:
                Neither a width nor a height was provided.

        Returns:
            str:
            The URL to the thumbnail, or ``None`` if one could not be
            generated.
        """
        if not width and not height:
            raise ValueError(
                _('Either a thumbnail width or height must be provided.'))

        return self.generate_thumbnail_image(width=width,
                                             height=height)

    def generate_thumbnail_image(
        self,
        *,
        width: Optional[int] = None,
        height: Optional[int] = None,
        create_if_missing: bool = True,
    ) -> Optional[str]:
        """Generate a thumbnail of a given size.

        For mimetype handlers that support image-based thumbnails, this must
        generate a thumbnail matching the given size requirements, and then
        return the URL to that thumbnail.

        The caller must provide a width or a height at minimum, and may provide
        both. If providing only a width or a height, the thumbnail size is
        expected to be constrained only to that dimension.

        Version Added:
            7.0

        Args:
            width (int, optional):
                The width to constrain the thumbnail image to.

            height (int, optional):
                The height to constrain the thumbnail image to.

            create_if_missing (bool, optional):
                Whether to create the thumbnail if one does not already exist.

                If ``False``, the existing thumbnail URL will be returned if
                it exists, but a new one will not otherwise be created.

        Raises:
            NotImplementedError:
                The mimetype handler does not support thumbnail images.

        Returns:
            str:
            The URL to the thumbnail, or ``None`` if one could not be
            generated.
        """
        raise NotImplementedError(
            _('%s does not support generating thumbnail images.')
            % type(self).__name__)

    def delete_associated_files(self) -> None:
        """Delete any extra files associated with this attachment.

        This should be implemented by subclasses who create and store extra
        files for file attachments, such as handlers that create and store
        thumbnail files. This should not delete the main file of the file
        attachment.

        Version Added:
            6.0
        """
        pass

    def _get_mimetype_file(self, name):
        return '%s/%s.png' % (self.MIMETYPES_DIR, name)


class ImageMimetype(MimetypeHandler):
    """Handles image mimetypes."""

    supported_mimetypes = ['image/*']

    def get_thumbnail(self):
        """Return a thumbnail of the image.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the thumbnail for the associated attachment.
        """
        attachment = self.attachment
        url = attachment.get_absolute_url()
        base_size = self.BASE_THUMBNAIL_IMAGE_WIDTH

        thumbnails = OrderedDict(
            (f'{scale}x', f'{url}?thumbnail=1&width={base_size * scale}')
            for scale in self.THUMBNAIL_IMAGE_SCALES
        )

        return format_html(
            '<div class="file-thumbnail">'
            ' <img src="{src}" srcset="{srcset}"'
            ' alt="{caption}" width="300" />'
            '</div>',
            src=thumbnails['1x'],
            srcset=', '.join(
                f'{_url} {_scale}'
                for _scale, _url in thumbnails.items()
            ),
            caption=attachment.caption)

    def generate_thumbnail_image(
        self,
        *,
        width: Optional[int] = None,
        height: Optional[int] = None,
        create_if_missing: bool = True,
    ) -> Optional[str]:
        """Generate a thumbnail of a given size.

        For mimetype handlers that support image-based thumbnails, this must
        generate a thumbnail matching the given size requirements, and then
        return the URL to that thumbnail.

        The caller must provide a width or a height at minimum, and may provide
        both. If providing only a width or a height, the thumbnail size is
        expected to be constrained only to that dimension.

        Version Added:
            7.0

        Args:
            width (int, optional):
                The width to constrain the thumbnail image to.

            height (int, optional):
                The height to constrain the thumbnail image to.

            create_if_missing (bool, optional):
                Whether to create the thumbnail if one does not already exist.

                If ``False``, the existing thumbnail URL will be returned if
                it exists, but a new one will not otherwise be created.

        Raises:
            NotImplementedError:
                The mimetype handler does not support thumbnail images.

        Returns:
            str:
            The URL to the thumbnail, or ``None`` if one could not be
            generated.
        """
        file = self.attachment.file

        if not file or not file.name:
            # This may occur during tests. There's no file contents saved,
            # so nothing to delete.
            return None

        return thumbnail(file,
                         size=(width, height),
                         create_if_missing=create_if_missing)

    def delete_associated_files(self) -> None:
        """Delete the thumbnail files for this attachment.

        Version Added:
            6.0
        """
        siteconfig = SiteConfiguration.objects.get_current()
        site_media_url = siteconfig.get('site_media_url')
        storage = self.attachment.file.storage

        base_size = self.BASE_THUMBNAIL_IMAGE_WIDTH

        for scale in self.THUMBNAIL_IMAGE_SCALES:
            url = self.generate_thumbnail_image(width=base_size * scale,
                                                create_if_missing=False)

            if not url:
                # There's no image to delete.
                continue

            filename: Optional[str] = None

            if url.startswith(site_media_url):
                filename = url[len(site_media_url):]

            if filename and storage.exists(filename):
                storage.delete(filename)


class TextMimetype(MimetypeHandler):
    """Handles text mimetypes.

    Text mimetypes provide thumbnails containing the first few lines of the
    file, syntax-highlighted.
    """

    supported_mimetypes = [
        'text/*',
        'application/javascript',
        'application/json',
        'application/x-javascript',
        'application/x-json',
        'application/x-yaml',
    ]

    # Read up to 'FILE_CROP_CHAR_LIMIT' number of characters from
    # the file attachment to prevent long reads caused by malicious
    # or auto-generated files.
    FILE_CROP_CHAR_LIMIT = 1000
    TEXT_CROP_NUM_HEIGHT = 50

    def _generate_preview_html(self, data):
        """Return the first few truncated lines of the text file.

        Args:
            data (bytes):
                The contents of the attachment.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML-safe thumbnail content.
        """
        from reviewboard.diffviewer.chunk_generator import \
            NoWrapperHtmlFormatter

        charset = self.mimetype[2].get('charset', 'ascii')

        try:
            text = data.decode(charset)
        except UnicodeDecodeError:
            logger.error('Could not decode text file attachment %s using '
                         'charset "%s"',
                         self.attachment.pk, charset)
            text = data.decode('utf-8', 'replace')

        try:
            lexer = guess_lexer_for_filename(self.attachment.filename, text)
        except ClassNotFound:
            lexer = TextLexer()

        lines = highlight(text, lexer, NoWrapperHtmlFormatter()).splitlines()

        return format_html_join(
            '',
            '<pre>{0}</pre>',
            (
                (mark_safe(line),)
                for line in lines[:self.TEXT_CROP_NUM_HEIGHT]
            ))

    def _generate_thumbnail(self):
        """Return the HTML for a thumbnail preview for a text file.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML-safe thumbnail content.
        """
        try:
            f = self.attachment.file.file
        except IOError as e:
            logger.error('Failed to locate file attachment %s: %s',
                         self.attachment.pk, e)
            return ''

        try:
            f.open()
            data = f.read(self.FILE_CROP_CHAR_LIMIT)
        except (ValueError, IOError) as e:
            logger.error('Failed to read from file attachment %s: %s',
                         self.attachment.pk, e)
            return ''
        finally:
            f.close()

        return format_html(
            '<div class="file-thumbnail">'
            ' <div class="file-thumbnail-clipped">{0}</div>'
            '</div>',
            self._generate_preview_html(data))

    def get_thumbnail(self):
        """Return the thumbnail of the text file as rendered as html.

        The content will be generated and then cached for future requests.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML-safe thumbnail content.
        """
        # Caches the generated thumbnail to eliminate the need on each page
        # reload to:
        # 1) re-read the file attachment
        # 2) re-generate the html based on the data read
        return mark_safe(
            cache_memoize('file-attachment-thumbnail-%s-html-%s'
                          % (self.__class__.__name__, self.attachment.pk),
                          self._generate_thumbnail))


class ReStructuredTextMimetype(TextMimetype):
    """Handles ReStructuredText (.rst) mimetypes.

    ReST mimetypes provide thumbnails containing the first few lines of
    rendered content from the file.
    """

    supported_mimetypes = ['text/x-rst', 'text/rst']

    def _generate_preview_html(self, data_string):
        """Return the HTML for a thumbnail preview for a ReST file.

        Args:
            data_string (bytes):
                The contents of the file.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML-safe thumbnail content.
        """
        # Use safe filtering against injection attacks
        docutils_settings = {
            'file_insertion_enabled': False,
            'raw_enabled': False,
            '_disable_config': True
        }

        parts = docutils.core.publish_parts(
            source=smart_str(data_string),
            writer_name='html4css1',
            settings_overrides=docutils_settings)

        return mark_safe(parts['html_body'])


class MarkDownMimetype(TextMimetype):
    """Handle MarkDown (.md) mimetypes.

    Markdown mimetypes provide thumbnails containing the first few lines of
    rendered content from the file.
    """

    supported_mimetypes = ['text/x-markdown', 'text/markdown']

    def _generate_preview_html(self, data_string):
        """Return the HTML for a thumbnail preview for a Markdown file.

        Returns:
            django.utils.safestring.SafeText:
            The resulting HTML-safe thumbnail content.
        """
        return mark_safe(render_markdown(force_str(data_string)))


class VideoMimetype(MimetypeHandler):
    """Handles video mimetypes.

    This will display a thumbnail utilizing the ``<video>`` tag, allowing a
    frame of video to be shown for browsers that support the video format.
    """

    supported_mimetypes = ['video/*']

    def get_thumbnail(self):
        """Return HTML that represents a preview of the attachment.

        This will create a ``<video>`` tag that starts half a second into
        the video, giving the browser a spot in which to load a frame for
        use in the thumbnail. The browser will fetch only what's needed in
        order to show this.

        Returns:
            django.utils.safestring.SafeText:
            The HTML for the video thumbnail.
        """
        attachment = self.attachment

        return format_html(
            '<div class="file-thumbnail">'
            '  <video width="300" preload="metadata" playsinline="true"'
            '         muted="true">'
            '  <source src="{src}#t=0.5" type="{mimetype}">'
            ' </video>'
            '</div>',
            src=attachment.get_absolute_url(),
            mimetype=attachment.mimetype)


# A mapping of mimetypes to icon names.
#
# Normally, a mimetype will be normalized and looked up in our bundled
# list of mimetype icons. However, if the mimetype is in this list, the
# associated name is used instead.
MIMETYPE_ICON_ALIASES = {
    'application/magicpoint': 'x-office-presentation',
    'application/msword': 'x-office-document',
    'application/ogg': 'audio-x-generic',
    'application/pdf': 'x-office-document',
    'application/postscript': 'x-office-document',
    'application/rtf': 'x-office-document',
    'application/vnd.lotus-1-2-3': 'x-office-spreadsheet',
    'application/vnd.ms-excel': 'x-office-spreadsheet',
    'application/vnd.ms-powerpoint': 'x-office-presentation',
    'application/vnd.oasis.opendocument.graphics': 'x-office-drawing',
    'application/vnd.oasis.opendocument.graphics-template':
        'x-office-drawing-template',
    'application/vnd.oasis.opendocument.image': 'x-office-drawing',
    'application/vnd.oasis.opendocument.presentation': 'x-office-presentation',
    'application/vnd.oasis.opendocument.presentation-template':
        'x-office-presentation-template',
    'application/vnd.oasis.opendocument.spreadsheet': 'x-office-spreadsheet',
    'application/vnd.oasis.opendocument.spreadsheet-template':
        'x-office-spreadsheet-template',
    'application/vnd.oasis.opendocument.text': 'x-office-document',
    'application/vnd.oasis.opendocument.text-template':
        'x-office-document-template',
    'application/vnd.oasis.opendocument.text-web': 'text-html',
    'application/vnd.rn-realmedia': 'video-x-generic',
    'application/vnd.rn-realmedia-secure': 'video-x-generic',
    'application/vnd.rn-realmedia-vbr': 'video-x-generic',
    'application/vnd.stardivision.calc': 'x-office-spreadsheet',
    'application/vnd.stardivision.impress': 'x-office-presentation',
    'application/vnd.stardivision.writer': 'x-office-document',
    'application/vnd.sun.xml.calc': 'x-office-spreadsheet',
    'application/vnd.sun.xml.calc.template': 'x-office-spreadsheet-template',
    'application/vnd.sun.xml.draw': 'x-office-drawing',
    'application/vnd.sun.xml.draw.template': 'x-office-drawing-template',
    'application/vnd.sun.xml.impress': 'x-office-presentation',
    'application/vnd.sun.xml.impress.template':
        'x-office-presentation-template',
    'application/vnd.sun.xml.writer': 'x-office-document',
    'application/vnd.sun.xml.writer.template': 'x-office-document-template',
    'application/wordperfect': 'x-office-document',
    'application/x-7z-compressed': 'package-x-generic',
    'application/x-abiword': 'x-office-document',
    'application/x-applix-spreadsheet': 'x-office-spreadsheet',
    'application/x-applix-word': 'x-office-document',
    'application/x-archive': 'package-x-generic',
    'application/x-arj': 'package-x-generic',
    'application/x-bzip-compressed-tar': 'package-x-generic',
    'application/x-bzip': 'package-x-generic',
    'application/x-compressed-tar': 'package-x-generic',
    'application/x-compress': 'package-x-generic',
    'application/x-cpio-compressed': 'package-x-generic',
    'application/x-cpio': 'package-x-generic',
    'application/x-deb': 'package-x-generic',
    'application/x-dvi': 'x-office-document',
    'application/x-executable': 'application-x-executable',
    'application/x-font-afm': 'font-x-generic',
    'application/x-font-bdf': 'font-x-generic',
    'application/x-font-linux-psf': 'font-x-generic',
    'application/x-font-pcf': 'font-x-generic',
    'application/x-font-sunos-news': 'font-x-generic',
    'application/x-font-ttf': 'font-x-generic',
    'application/x-gnumeric': 'x-office-spreadsheet',
    'application/x-gzip': 'package-x-generic',
    'application/gzip': 'package-x-generic',
    'application/x-gzpostscript': 'x-office-document',
    'application/xhtml+xml': 'text-html',
    'application/x-jar': 'package-x-generic',
    'application/x-killustrator': 'image-x-generic',
    'application/x-kpresenter': 'x-office-presentation',
    'application/x-kspread': 'x-office-spreadsheet',
    'application/x-kword': 'x-office-document',
    'application/x-lha': 'package-x-generic',
    'application/x-lhz': 'package-x-generic',
    'application/x-lzma-compressed-tar': 'package-x-generic',
    'application/x-lzma': 'package-x-generic',
    'application/x-ms-dos-executable': 'application-x-executable',
    'application/x-perl': 'text-x-script',
    'application/x-php': 'text-html',
    'application/x-python-bytecode': 'text-x-script',
    'application/x-rar': 'package-x-generic',
    'application/x-rpm': 'package-x-generic',
    'application/x-scribus': 'x-office-document',
    'application/x-shellscript': 'text-x-script',
    'application/x-shockwave-flash': 'video-x-generic',
    'application/x-stuffit': 'package-x-generic',
    'application/x-tar': 'package-x-generic',
    'application/x-tarz': 'package-x-generic',
    'application/x-tex': 'x-office-document',
    'application/zip': 'package-x-generic',
    'text/html': 'text-html',
    'text/vnd.wap.wml': 'text-html',
    'text/x-csh': 'text-x-script',
    'text/x-python': 'text-x-script',
    'text/x-sh': 'text-x-script',
    'text/x-vcalendar': 'x-office-calendar',
    'text/x-vcard': 'x-office-address-book',
    'text/x-zsh': 'text-x-script',
}


# A mapping of file extensions to mimetypes
#
# Normally mimetypes are determined by mimeparse, then matched with
# one of the supported mimetypes classes through a best-match algorithm.
# However, mimeparse isn't always able to catch the unofficial mimetypes
# such as 'text/x-rst' or 'text/x-markdown', so we just go by the
# extension name.
MIMETYPE_EXTENSIONS = {
    '.rst': ('text', 'x-rst', {}),
    '.md': ('text', 'x-markdown', {}),
}



"""Move from Meta.index_together to Meta.indexes for ReviewRequestVisit."""

from django_evolution.mutations import ChangeMeta


MUTATIONS = [
    ChangeMeta('ReviewRequestVisit', 'indexes',
               [{'fields': ['user', 'visibility']}]),
    ChangeMeta('ReviewRequestVisit', 'index_together', []),
]
